"""

Deployment descriptor API.

Deployment descriptors live in the deployment/descriptors directory. They are
YAML-formatted files. Each descriptor has three things in it:

1. A field for the Capo profile containing the main database connection for
   the archive

2. One or more categories, such as web, workflow, frontend, and solr, which
   contains a list of hosts.

3. A dictionary of host to the Capo profile that pertains to that host.

---

Deployment descriptors are parsed by calling descriptor.load(descriptor name).
A descriptor name is something like "test" or "production"; the loader will
look in descriptors/name.yaml for the file.

The overall deployment descriptor is represented by the DeploymentDescriptor
class. That class provides properties for each category, and can find the
primary database settings.

Each category holds a list of hosts.

Each host can produce an ssh Connection, or settings or Capo properties, as well as
supporting a WAR file deployment method.

"""

import deployment.tomcat
import yaml
from fabric import Connection
from pycapo import CapoConfig


class Host:
    """
    A host for deploying things to.
    """
    def __init__(self, category, host):
        self.category = category
        self.host = host

    def __repr__(self):
        return f'<Host {self.host}>'

    @property
    def capo_profile(self):
        """
        :return: the name of the Capo profile that pertains to this host
        """
        return self.category.capo_profile_for(self.host)

    @property
    def capo(self):
        """
        :return: the CapoConfig for this host
        """
        return CapoConfig(profile=self.capo_profile)

    def settings(self, prefix):
        """
        Returns the Settings for this prefix on this host.
        :param prefix: the settings prefix, like "tomcat.deployment"
        :return: the Capo Settings object for the given prefix, in this profile for this host.
        """
        return self.capo.settings(prefix)

    def tomcat_credentials(self):
        """
        :return: a tuple of (username, password) for the Tomcat manager on this host
        """
        settings = self.settings('tomcat.deployment')
        return settings.username, settings.password

    def war_deploy(self, c, warfile):
        """
        Deploy a WAR file to Tomcat on this host
        :param c:        this is the invoke context
        :param warfile:  the war file to deploy
        :return:         None
        """
        print(f'Deploying {warfile} to {self.host}')
        username, password = self.tomcat_credentials()
        deployment.tomcat.war(c, self.host, warfile, username=username, password=password)

    def connection(self):
        """
        :return: an SSH connection to this host.
        """
        return Connection(self.host)


class Category:
    """
    A category of hosts, such as machines that run the workflow,
    or machines that run the front-end.
    """
    def __init__(self, dd, category, hosts):
        self.dd = dd
        self.category = category
        self.hosts = hosts

    def __repr__(self):
        return f'<Category {self.category} with {len(self.hosts)} hosts>'

    def __iter__(self):
        """
        The main way of using this class: as a generator of Hosts.
        :return:  a generator of Host instances in this category
        """
        return (Host(self, host) for host in self.hosts)

    def capo_profile_for(self, host):
        """
        Look up in the deployment descriptor the Capo profile for this host.
        :param host:  the host to look up
        :return:      the name of the Capo profile that pertains to it
        """
        return self.dd.capo_profile_for(host)


class DeploymentDescriptor:
    """
    Abstracts a deployment descriptor, providing categories of hosts and
    the overall database connection, plus providing support methods
    used by the other classes in this package.

    None of the other classes have direct access to the YAML, so if the
    shape of the YAML changes, only this class will need to be modified.
    """
    def __init__(self, descriptor):
        self.descriptor = descriptor
        self.yaml = yaml.load(open(f'deployment/descriptors/{descriptor}.yaml'))

    def __getattr__(self, category):
        """
        Main way of using this class: as a bunch of category properties
        :param category:   the category to look up
        :return:           the Category instance for this name
        """
        return Category(self, category, self.yaml[category])

    def __repr__(self):
        return f'<DeploymentDescriptor {self.descriptor}>'

    def capo_profile_for(self, host):
        """
        Look up the Capo profile for this host in our YAML
        :param host:   the host to look up
        :return:       the Capo profile for this host
        """
        return self.yaml['capo_profiles'][host]

    def database_capo_profile(self):
        return self.yaml['database']

    def database_settings(self):
        """
        :return: the metadata database settings for this deployment
        """
        return CapoConfig(profile=self.database_capo_profile()).settings('metadataDatabase')


def load(descriptor):
    return DeploymentDescriptor(descriptor)


